# SPDX-FileCopyrightText: 2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'idf-build-apps'
project_homepage = 'https://github.com/espressif/idf-build-apps'
copyright = '2023, Espressif Systems (Shanghai) Co., Ltd.'
author = 'Fu Hanxi'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx_copybutton',
    'myst_parser',
    'sphinxcontrib.mermaid',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_css_files = ['theme_overrides.css']
html_logo = '_static/espressif-logo.svg'
html_static_path = ['_static']
html_theme = 'sphinx_rtd_theme'

# mermaid 10.0.0 change the CDN path
# temp workaround until sphinxcontrib-mermaid fixed it
mermaid_version = '9.4.0'



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import pytest

from idf_build_apps import (
    App,
)


@pytest.fixture(autouse=True)
def clean_cls_attr():
    App.MANIFEST = None



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

from pathlib import (
    Path,
)

import pytest

from idf_build_apps.utils import (
    files_matches_patterns,
    get_parallel_start_stop,
    rmdir,
)


@pytest.mark.parametrize(
    'patterns, expected',
    [
        ('*.txt', ['test/inner', 'test/inner/test.txt', 'test/test.txt']),
        (
            ['*.txt', '*.log'],
            [
                'test/inner',
                'test/inner/build.log',
                'test/inner/test.txt',
                'test/test.txt',
            ],
        ),
    ],
)
def test_rmdir(tmpdir, patterns, expected):
    test_dir = tmpdir.mkdir('test')
    dir1 = test_dir.mkdir('inner')
    test_dir.mkdir('inner2')

    Path(dir1 / 'test.txt').touch()
    Path(dir1 / 'build.log').touch()
    Path(test_dir / 'test.txt').touch()

    rmdir(test_dir, exclude_file_patterns=patterns)

    assert sorted(Path(test_dir).glob('**/*')) == [Path(tmpdir / i) for i in expected]


@pytest.mark.parametrize(
    'total, parallel_count, parallel_index, start, stop',
    [
        (1, 1, 1, 1, 1),
        (1, 2, 2, 2, 1),
        (1, 10, 1, 1, 1),
        (6, 4, 1, 1, 2),
        (6, 4, 2, 3, 4),
        (6, 4, 3, 5, 6),
        (6, 4, 4, 7, 6),
        (10, 10, 9, 9, 9),
        (33, 2, 1, 1, 17),
        (33, 2, 2, 18, 33),
    ],
)
def test_get_parallel_start_stop(total, parallel_count, parallel_index, start, stop):
    assert (start, stop) == get_parallel_start_stop(total, parallel_count, parallel_index)


@pytest.mark.parametrize(
    'files, patterns, rootpath, result',
    [
        ('c.py', '*.py', None, True),
        ('c.py', '*.py', '/a', True),
        ('/a/b/c.py', 'c.py', None, False),
        ('/a/b/c.py', 'c.py', '/a/b', True),
        ('/a/b/../b/c.py', '*.py', None, False),
        ('/a/b/../b/c.py', '*.py', '/a/b', True),
        ('/a/b/../b/c.py', '**/*.py', None, False),
        ('/a/b/../b/c.py', '**/*.py', '/a', True),
        ('c.py', '/a/**/*.py', None, False),
        ('c.py', '/a/**/*.py', '/a', False),
        ('c.py', '/a/**/*.py', '/a/b', True),
        ('/a/b/c.py', '/a/**/*.py', None, True),
        ('/a/b/c.py', '/a/**/*.py', 'foo', True),
    ],
)
def test_files_matches_patterns(files, patterns, rootpath, result):
    assert files_matches_patterns(files, patterns, rootpath) == result



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

"""
Tools for building ESP-IDF related apps.
"""

import logging

__version__ = '0.4.1'

LOGGER = logging.getLogger('idf_build_apps')

from .app import (
    App,
    CMakeApp,
)
from .main import (
    build_apps,
    find_apps,
)
from .utils import (
    setup_logging,
)

__all__ = [
    'App',
    'CMakeApp',
    'find_apps',
    'build_apps',
    'setup_logging',
]



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os
import re
import shutil
from pathlib import (
    Path,
)

from . import (
    LOGGER,
)
from .app import (
    App,
)
from .constants import (
    ALL_TARGETS,
)
from .finder import (
    _find_apps,
)
from .manifest.manifest import (
    FolderRule,
    Manifest,
)
from .utils import (
    BuildError,
    files_matches_patterns,
    get_parallel_start_stop,
    to_list,
)

try:
    import typing as t
except ImportError:
    pass


def find_apps(
    paths,  # type: list[str] | str
    target,  # type: str
    build_system='cmake',  # type: str
    recursive=False,  # type: bool
    exclude_list=None,  # type: list[str] | None
    work_dir=None,  # type: str | None
    build_dir='build',  # type: str
    config_rules_str=None,  # type: list[str] | None
    build_log_path=None,  # type: str | None
    size_json_path=None,  # type: str | None
    check_warnings=False,  # type: bool
    preserve=True,  # type: bool
    manifest_files=None,  # type: list[str] | str | None
    default_build_targets=None,  # type: list[str] | str | None
    depends_on_components=None,  # type: list[str] | str | None
    manifest_rootpath=None,  # type: str | None
    ignore_component_dependencies_file_patterns=None,  # type: list[str] | str | None
    depends_on_files=None,  # type: list[str] | str | None
):  # type: (...) -> list[App]
    """
    Find app directories in paths (possibly recursively), which contain apps for the given build system, compatible
    with the given target

    :param paths: list of app directories (can be / usually will be a relative path)
    :type paths: list[str] | str
    :param target: desired value of IDF_TARGET; apps incompatible with the given target are skipped.
    :type target: str
    :param build_system: name of the build system, now only support cmake
    :type build_system: str
    :param recursive: Recursively search into the nested sub-folders if no app is found or not
    :type recursive: bool
    :param exclude_list: list of paths to be excluded from the recursive search
    :type exclude_list: list[str] | None
    :param work_dir: directory where the app should be copied before building. Support placeholders
    :type work_dir: str | None
    :param build_dir: directory where the build will be done. Support placeholders.
    :type build_dir: str
    :param config_rules_str: mapping of sdkconfig file name patterns to configuration names
    :type config_rules_str: list[str] | None
    :param build_log_path: path of the build log. Support placeholders.
        The logs will go to stdout/stderr if not specified
    :type build_log_path: str | None
    :param size_json_path: path of the size.json file. Support placeholders.
        Will not generate size file for each app if not specified
    :type size_json_path: str | None
    :param check_warnings: Check for warnings in the build log or not
    :type check_warnings: bool
    :param preserve: Preserve the built binaries or not
    :type preserve: bool
    :param manifest_files: paths of the manifest files
    :type manifest_files: list[str] | str | None
    :param default_build_targets: default build targets used in manifest files
    :type default_build_targets: list[str] | str | None
    :param depends_on_components: app with ``requires_components`` set in the corresponding manifest files will only
        be built if it depends on any of the specified components
    :type depends_on_components: list[str] | str | None
    :param manifest_rootpath: The root path of the manifest files. Usually the folders specified in the manifest files
        are relative paths. Use the current directory if not specified
    :type manifest_rootpath: str | None
    :param ignore_component_dependencies_file_patterns: file patterns that use to ignore checking the component
        dependencies
    :type ignore_component_dependencies_file_patterns: list[str] | str | None
    :param depends_on_files: skip check app's component dependencies if any of the specified files matches
        ``ignore_component_dependencies_file_patterns``
    :type depends_on_files: list[str] | str | None
    :return: list of found apps
    :rtype: list[App]
    """
    if default_build_targets:
        default_build_targets = to_list(default_build_targets)
        LOGGER.info('Overriding DEFAULT_BUILD_TARGETS to %s', default_build_targets)
        FolderRule.DEFAULT_BUILD_TARGETS = default_build_targets

    # always set the manifest rootpath at the very beginning of find_apps in case ESP-IDF switches the branch.
    Manifest.ROOTPATH = Path(manifest_rootpath or os.curdir).resolve()

    if manifest_files:
        rules = set()
        for _manifest_file in to_list(manifest_files):
            LOGGER.debug('Loading manifest file: %s', _manifest_file)
            rules.update(Manifest.from_file(_manifest_file).rules)
        manifest = Manifest(rules)
        App.MANIFEST = manifest

    depends_on_components = to_list(depends_on_components)
    if depends_on_components is None:
        check_component_dependencies = False
    elif (
        ignore_component_dependencies_file_patterns
        and depends_on_files
        and files_matches_patterns(depends_on_files, ignore_component_dependencies_file_patterns)
    ):
        LOGGER.debug(
            'Skipping check component dependencies for apps since files %s matches patterns: %s',
            ', '.join(depends_on_files),
            ', '.join(ignore_component_dependencies_file_patterns),
        )
        check_component_dependencies = False
    else:
        check_component_dependencies = True

    apps = []
    if target == 'all':
        targets = ALL_TARGETS
    else:
        targets = [target]

    for target in targets:
        for path in to_list(paths):
            apps.extend(
                _find_apps(
                    path,
                    target,
                    build_system,
                    recursive,
                    exclude_list or [],
                    work_dir=work_dir,
                    build_dir=build_dir or 'build',
                    config_rules_str=config_rules_str,
                    build_log_path=build_log_path,
                    size_json_path=size_json_path,
                    check_warnings=check_warnings,
                    preserve=preserve,
                    depends_on_components=depends_on_components,
                    check_component_dependencies=check_component_dependencies,
                )
            )
    apps.sort()

    LOGGER.info('Found %d apps in total', len(apps))
    return apps


def build_apps(
    apps,  # type: list[App] | App
    build_verbose=False,  # type: bool
    parallel_count=1,  # type: int
    parallel_index=1,  # type: int
    dry_run=False,  # type: bool
    keep_going=False,  # type: bool
    collect_size_info=None,  # type: t.TextIO | None
    collect_app_info=None,  # type: t.TextIO | None
    ignore_warning_strs=None,  # type: list[str] | None
    ignore_warning_file=None,  # type: t.TextIO | None
    copy_sdkconfig=False,  # type: bool
    depends_on_components=None,  # type: list[str] | str | None
    manifest_rootpath=None,  # type: str | None
    ignore_component_dependencies_file_patterns=None,  # type: list[str] | str | None
    depends_on_files=None,  # type: list[str] | str | None
):  # type: (...) -> (int, list[App]) | int
    """
    Build all the specified apps

    :param apps: list of apps to be built
    :type apps: list[App] | App
    :param build_verbose: call ``--verbose`` in ``idf.py build`` or not
    :type build_verbose: bool
    :param parallel_count: number of parallel tasks to run
    :type parallel_count: int
    :param parallel_index: index of the parallel task to run
    :type parallel_index: int
    :param dry_run: simulate this run or not
    :type dry_run: bool
    :param keep_going: keep building or not if one app's build failed
    :type keep_going: bool
    :param collect_size_info: file path to record all generated size files' paths if specified
    :type collect_size_info: TextIO | None
    :param collect_app_info: file path to record all the built apps' info if specified
    :type collect_app_info: TextIO | None
    :param ignore_warning_strs: ignore build warnings that matches any of the specified regex patterns
    :type ignore_warning_strs: list[str] | None
    :param ignore_warning_file: ignore build warnings that matches any of the lines of the regex patterns in the
        specified file
    :type ignore_warning_file: list[str] | None
    :param copy_sdkconfig: copy the sdkconfig file to the build directory or not
    :type copy_sdkconfig: bool
    :param depends_on_components: app with ``requires_components`` set in the corresponding manifest files would only be
        built if it depends on any of the specified components
    :type depends_on_components: list[str] | str | None
    :param manifest_rootpath: The root path of the manifest files. Usually the folders specified in the manifest files
        are relative paths. Use the current directory if not specified
    :type manifest_rootpath: str | None
    :param ignore_component_dependencies_file_patterns: file patterns that use to ignore checking the component
        dependencies
    :type ignore_component_dependencies_file_patterns: list[str] | str | None
    :param depends_on_files: skip check app's component dependencies if any of the specified files matches
        ``ignore_component_dependencies_file_patterns``
    :type depends_on_files: list[str] | str | None
    :return: exit_code, built_apps if specified ``depends_on_components``
    :rtype: int, list[App]
    :return: exit_code if not specified ``depends_on_components``
    :rtype: int
    """
    apps = to_list(apps)

    ignore_warnings_regexes = []
    if ignore_warning_strs:
        for s in ignore_warning_strs:
            ignore_warnings_regexes.append(re.compile(s))
    if ignore_warning_file:
        for s in ignore_warning_file:
            ignore_warnings_regexes.append(re.compile(s.strip()))
    App.IGNORE_WARNS_REGEXES = ignore_warnings_regexes

    depends_on_components = to_list(depends_on_components)
    if depends_on_components is None:
        check_component_dependencies = False
    elif (
        ignore_component_dependencies_file_patterns
        and depends_on_files
        and files_matches_patterns(depends_on_files, ignore_component_dependencies_file_patterns, manifest_rootpath)
    ):
        LOGGER.debug(
            'Skipping check component dependencies for apps since files %s matches patterns: %s',
            ', '.join(depends_on_files),
            ', '.join(ignore_component_dependencies_file_patterns),
        )
        check_component_dependencies = False
    else:
        check_component_dependencies = True

    start, stop = get_parallel_start_stop(len(apps), parallel_count, parallel_index)
    LOGGER.info('Total %s apps. running build for app %s-%s', len(apps), start, stop)

    failed_apps = []
    exit_code = 0

    LOGGER.info('Building the following apps:')
    if apps[start - 1 : stop]:
        for app in apps[start - 1 : stop]:
            LOGGER.info('  %s (preserve: %s)', app, app.preserve)
    else:
        LOGGER.info('  parallel count is too large. build nothing...')

    actual_built_apps = []
    for i, app in enumerate(apps):
        index = i + 1  # we use 1-based
        if index < start or index > stop:
            continue

        # attrs
        app.dry_run = dry_run
        app.index = index
        app.verbose = build_verbose

        LOGGER.info('Building app %s: %s', index, repr(app))
        is_built = False
        try:
            is_built = app.build(
                depends_on_components=depends_on_components,
                check_component_dependencies=check_component_dependencies,
            )
        except BuildError as e:
            LOGGER.error(str(e))
            if keep_going:
                failed_apps.append(app)
                exit_code = 1
            else:
                if depends_on_components is not None:
                    return 1, actual_built_apps
                else:
                    return 1
        finally:
            if is_built:
                actual_built_apps.append(app)

                if collect_app_info:
                    collect_app_info.write(app.to_json() + '\n')
                    LOGGER.info('=> Recorded app info in %s', collect_app_info.name)

                if collect_size_info:
                    try:
                        app.collect_size_info(collect_size_info)
                    except Exception as e:
                        LOGGER.warning('Adding size info for app %s failed:', app.name)
                        LOGGER.warning(e)
                        pass

                if copy_sdkconfig:
                    try:
                        shutil.copy(
                            os.path.join(app.work_dir, 'sdkconfig'),
                            os.path.join(app.build_path, 'sdkconfig'),
                        )
                    except Exception as e:
                        LOGGER.warning('Copy sdkconfig file from app %s work dir %s failed:', app.name, app.work_dir)
                        LOGGER.warning(e)
                        pass
                    else:
                        LOGGER.info('=> Copied sdkconfig file from %s to %s', app.work_dir, app.build_path)

            LOGGER.info('')  # add one empty line for separating different builds

    if failed_apps:
        LOGGER.error('Build failed for the following apps:')
        for app in failed_apps:
            LOGGER.error('  %s', app)

    if depends_on_components is not None:
        return exit_code, actual_built_apps
    else:
        return exit_code



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import inspect
import logging
import os

import pytest

from idf_build_apps.constants import (
    DEFAULT_SDKCONFIG,
    IDF_PATH,
)
from idf_build_apps.main import (
    find_apps,
)


def test_finder(tmpdir):
    test_dir = str(IDF_PATH / 'examples')
    apps = find_apps(test_dir, 'esp32', recursive=True)
    assert apps

    yaml_file = str(tmpdir / 'test.yml')
    with open(yaml_file, 'w') as fw:
        fw.write(
            inspect.cleandoc(
                '''
            {}:
                enable:
                    - if: IDF_TARGET == "esp32s2"
        '''.format(
                    test_dir
                )
            )
        )
    filtered_apps = find_apps(test_dir, 'esp32', recursive=True, manifest_files=yaml_file)
    assert not filtered_apps
    assert filtered_apps != apps


def test_finder_with_sdkconfig_defaults():
    test_dir = str(IDF_PATH / 'examples')
    apps = find_apps(test_dir, 'esp32', recursive=True)
    assert apps

    # write the first app without sdkconfig.defaults with CONFIG_IDF_TARGET="linux" in sdkconfig.defaults
    _app = None
    _default_sdkconfig_path = None
    for app in apps:
        default_sdkconfig_path = os.path.join(app.app_dir, DEFAULT_SDKCONFIG)
        if not os.path.isfile(default_sdkconfig_path):
            with open(default_sdkconfig_path, 'w') as fw:
                fw.write('CONFIG_IDF_TARGET="linux"')
                logging.info('Created temp %s %s', DEFAULT_SDKCONFIG, _default_sdkconfig_path)
            _app = app
            _default_sdkconfig_path = default_sdkconfig_path
            break
    else:
        raise ValueError('no app without {}'.format(DEFAULT_SDKCONFIG))

    try:
        filtered_apps = find_apps(test_dir, 'esp32', recursive=True)
        assert set(apps) - set(filtered_apps) == {_app}
    finally:
        try:
            os.remove(_default_sdkconfig_path)
            logging.info('Removed temp %s %s', DEFAULT_SDKCONFIG, _default_sdkconfig_path)
        except:  # noqa
            pass


@pytest.mark.parametrize(
    'depends_on_components, could_find_apps',
    [
        (None, True),
        ([], False),
        ('fake', False),
        ('soc', True),
        (['soc', 'fake'], True),
    ],
)
def test_finder_with_requires_and_depends_on_components(tmpdir, depends_on_components, could_find_apps):
    test_dir = str(IDF_PATH / 'examples')
    apps = find_apps(test_dir, 'esp32', recursive=True)
    assert apps

    yaml_file = str(tmpdir / 'test.yml')
    with open(yaml_file, 'w') as fw:
        fw.write(
            inspect.cleandoc(
                '''
            {}:
                requires_components:
                    - freertos
                    - soc
        '''.format(
                    test_dir
                )
            )
        )

    filtered_apps = find_apps(
        test_dir,
        'esp32',
        recursive=True,
        manifest_files=yaml_file,
        depends_on_components=depends_on_components,
    )
    if could_find_apps:
        assert filtered_apps == apps
    else:
        assert not filtered_apps


def test_finder_after_chdir():
    test_dir = IDF_PATH / 'examples' / 'get-started'

    yaml_file = str(test_dir / 'test.yml')
    with open(yaml_file, 'w') as fw:
        fw.write(
            inspect.cleandoc(
                '''
            examples/get-started:
                enable:
                    - if: IDF_TARGET != "esp32"
        '''
            )
        )

    os.chdir(IDF_PATH)
    assert not find_apps(str(test_dir), 'esp32', recursive=True, manifest_files=yaml_file)

    # manifest folder invalid
    os.chdir(test_dir)
    assert find_apps(str(test_dir), 'esp32', recursive=True, manifest_files=yaml_file)


def test_finder_custom_root_dir():
    test_dir = IDF_PATH / 'examples' / 'get-started'

    yaml_file = str(test_dir / 'test.yml')
    with open(yaml_file, 'w') as fw:
        fw.write(
            inspect.cleandoc(
                '''
            get-started:
                enable:
                    - if: IDF_TARGET != "esp32"
        '''
            )
        )

    assert find_apps(str(test_dir), 'esp32', recursive=True, manifest_files=yaml_file, manifest_rootpath=str(IDF_PATH))

    assert not find_apps(
        str(test_dir), 'esp32', recursive=True, manifest_files=yaml_file, manifest_rootpath=str(IDF_PATH / 'examples')
    )


def test_finder_idf_version():
    test_dir = IDF_PATH / 'examples' / 'get-started'
    apps = find_apps(str(test_dir), 'esp32', recursive=True)
    assert apps

    yaml_file = str(test_dir / 'test.yml')
    with open(yaml_file, 'w') as fw:
        fw.write(
            inspect.cleandoc(
                '''
            get-started:
                enable:
                    - if: IDF_VERSION_MAJOR > 0 and IDF_VERSION_MINOR < 999 and IDF_VERSION_PATCH in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        '''
            )
        )

    assert find_apps(str(test_dir), 'esp32', recursive=True, manifest_files=yaml_file) == apps



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import argparse
import sys

from .constants import (
    ALL_TARGETS,
)
from .main import (
    build_apps,
    find_apps,
)
from .utils import (
    InvalidCommand,
    setup_logging,
)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Tools for building ESP-IDF related apps.'
        'Some CLI options can be expanded by the following placeholders, like "--work-dir", "--build-dir", etc.:\n'
        '- @t: would be replaced by the target chip type\n'
        '- @w: would be replaced by the wildcard, usually the sdkconfig\n'
        '- @n: would be replaced by the app name\n'
        '- @f: would be replaced by the escaped app path (replaced "/" to "_")\n'
        '- @i: would be replaced by the build index',
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    actions = parser.add_subparsers(dest='action')

    common_args = argparse.ArgumentParser(add_help=False)
    common_args.add_argument('-p', '--paths', nargs='+', help='One or more paths to look for apps.')
    common_args.add_argument('-t', '--target', help='filter apps by given target.')
    common_args.add_argument(
        '--build-system',
        default='cmake',
        choices=['cmake'],
        help='build with given build system',
    )
    common_args.add_argument(
        '--recursive',
        action='store_true',
        help='Look for apps in the specified directories recursively.',
    )
    common_args.add_argument(
        '--exclude',
        action='append',
        help='Ignore specified directory (if --recursive is given). Can be used multiple times.',
    )
    common_args.add_argument(
        '--work-dir',
        help='If set, the app is first copied into the specified directory, and then built. '
        'If not set, the work directory is the directory of the app. Can expand placeholders.',
    )
    common_args.add_argument(
        '--build-dir',
        help='If set, specifies the build directory name. Can expand placeholders. Can be either a '
        'name relative to the work directory, or an absolute path.',
    )
    common_args.add_argument(
        '--build-log',
        help='If specified, the build log will be written to this file. Can expand placeholders.',
    )
    common_args.add_argument(
        '--size-file',
        help='the size json will be written to this file. Can expand placeholders.',
    )
    common_args.add_argument(
        '--config',
        action='append',
        help='Adds configurations (sdkconfig file names) to build. Could be specified for multiple times.'
        'This can either be '
        'FILENAME[=NAME] or FILEPATTERN. FILENAME is the name of the sdkconfig file, '
        'relative to the project directory, to be used. Optional NAME can be specified, '
        'which can be used as a name of this configuration. FILEPATTERN is the name of '
        'the sdkconfig file, relative to the project directory, with at most one wildcard. '
        'The part captured by the wildcard is used as the name of the configuration.',
    )
    common_args.add_argument(
        '-v',
        '--verbose',
        default=0,
        action='count',
        help='Increase the logging level of the whole process. Can be specified multiple times.',
    )
    common_args.add_argument(
        '--log-file',
        type=argparse.FileType('w'),
        help='Write the script log to the specified file, instead of stderr',
    )
    common_args.add_argument(
        '--check-warnings',
        action='store_true',
        help='Check for warnings in the build output.',
    )
    common_args.add_argument(
        '--manifest-file',
        action='append',
        help='manifest file to specify the build test rules of the apps, could be specified multiple times.',
    )
    common_args.add_argument(
        '--manifest-rootpath',
        help='Root directory for calculating the realpath of the relative path defined in the manifest files. '
        'Would use the current directory if not set.',
    )
    common_args.add_argument(
        '--default-build-targets',
        help='comma-separated target list. IDF supported targets would be used if this option is not set',
    )
    common_args.add_argument(
        '--depends-on-components',
        nargs='*',
        default=None,
        help='space-separated components list, app with `requires_components` set in the corresponding manifest files '
        'would only be built if depends on any of the specified components',
    )
    common_args.add_argument(
        '-if',
        '--ignore-component-dependencies-file-patterns',
        nargs='*',
        default=None,
        help='ignore component dependencies when changed files matches any of the specified file patterns. must used '
        'with --depends-on-files',
    )
    common_args.add_argument(
        '--depends-on-files',
        nargs='*',
        default=None,
        help='space-separated file pattern list, the `requires_components` set in the manifest files will be '
        'ignored when specified files match any of the specified file patterns defined with '
        '--ignore-component-dependencies-file-patterns. Must used with --ignore-component-dependencies-file-patterns',
    )
    common_args.add_argument(
        '--no-color',
        action='store_true',
        help='enable colored output by default on UNIX-like systems. enable this flag to make the logs uncolored.',
    )

    find_parser = actions.add_parser('find', parents=[common_args])
    find_parser.add_argument(
        '-o',
        '--output',
        help='Output the found apps to the specified file instead of sys.stdout.',
    )

    build_parser = actions.add_parser('build', parents=[common_args])
    build_parser.add_argument(
        '--build-verbose',
        action='store_true',
        help='Enable verbose output from build system.',
    )
    build_parser.add_argument(
        '--parallel-count',
        default=1,
        type=int,
        help="Number of parallel build jobs. Note that this script doesn't start the jobs, "
        + 'it needs to be executed multiple times with same value of --parallel-count and '
        + 'different values of --parallel-index.',
    )
    build_parser.add_argument(
        '--parallel-index',
        default=1,
        type=int,
        help='Index (1-based) of the job, out of the number specified by --parallel-count.',
    )
    build_parser.add_argument(
        '--dry-run',
        action='store_true',
        help="Don't actually build, only print the build commands",
    )
    build_parser.add_argument(
        '--keep-going',
        action='store_true',
        help="Don't exit immediately when a build fails.",
    )
    build_parser.add_argument(
        '--no-preserve',
        action='store_true',
        help="Don't preserve the build directory after a successful build.",
    )
    build_parser.add_argument(
        '--collect-size-info',
        type=argparse.FileType('w'),
        help='write size info json file while building into the specified file. each line is a json object.',
    )
    build_parser.add_argument(
        '--collect-app-info',
        type=argparse.FileType('w'),
        help='write app info json file while building into the specified file. each line is a json object.',
    )
    build_parser.add_argument(
        '--ignore-warning-str',
        action='append',
        help='Ignore the warning string that match the specified regex in the build output. '
        'Can be specified multiple times.',
    )
    build_parser.add_argument(
        '--ignore-warning-file',
        type=argparse.FileType('r'),
        help='Ignore the warning strings in the specified file. Each line should be a regex string.',
    )
    build_parser.add_argument(
        '--copy-sdkconfig',
        action='store_true',
        help='Copy the sdkconfig file to the build directory.',
    )

    args = parser.parse_args()
    if args.action not in ['find', 'build']:
        parser.print_help()
        raise InvalidCommand('subcommand is required. {find, build}')

    if not args.paths:
        if args.action == 'find':
            find_parser.print_help()
        elif args.action == 'build':
            build_parser.print_help()

        raise InvalidCommand('Must specify at least one search path with CLI option "-p <path>" or "--path <path>"')

    setup_logging(args.verbose, args.log_file, not args.no_color)

    default_build_targets = []
    if args.default_build_targets:
        for t in args.default_build_targets.split(','):
            t = t.strip()
            if t not in ALL_TARGETS:
                print('Unrecognizable target {}, only know targets {}'.format(t, ALL_TARGETS))
                sys.exit(1)

            if t not in default_build_targets:
                default_build_targets.append(t)

    if (args.ignore_component_dependencies_file_patterns is None) != (args.depends_on_files is None):
        raise InvalidCommand(
            'Must specify "--ignore-component-dependencies-file-patterns" and "--depends-on-files" together'
        )

    apps = find_apps(
        args.paths,
        args.target,
        build_system=args.build_system,
        recursive=args.recursive,
        exclude_list=args.exclude or [],
        work_dir=args.work_dir,
        build_dir=args.build_dir or 'build',
        config_rules_str=args.config,
        build_log_path=args.build_log,
        size_json_path=args.size_file,
        check_warnings=args.check_warnings,
        manifest_files=args.manifest_file,
        default_build_targets=default_build_targets,
        depends_on_components=args.depends_on_components,
        manifest_rootpath=args.manifest_rootpath,
        ignore_component_dependencies_file_patterns=args.ignore_component_dependencies_file_patterns,
        depends_on_files=args.depends_on_files,
    )

    if args.action == 'find':
        if args.output:
            with open(args.output, 'w') as f:
                for app in apps:
                    f.write(str(app) + '\n')
        else:
            print('\n'.join([str(app) for app in apps]))

        sys.exit(0)

    # build from now on
    if args.no_preserve:
        for app in apps:
            app.preserve = False

    res = build_apps(
        apps,
        build_verbose=args.build_verbose,
        parallel_count=args.parallel_count,
        parallel_index=args.parallel_index,
        dry_run=args.dry_run,
        keep_going=args.keep_going,
        collect_size_info=args.collect_size_info,
        collect_app_info=args.collect_app_info,
        ignore_warning_strs=args.ignore_warning_str,
        ignore_warning_file=args.ignore_warning_file,
        copy_sdkconfig=args.copy_sdkconfig,
        depends_on_components=args.depends_on_components,
        manifest_rootpath=args.manifest_rootpath,
        ignore_component_dependencies_file_patterns=args.ignore_component_dependencies_file_patterns,
        depends_on_files=args.depends_on_files,
    )

    if args.depends_on_components is not None:
        sys.exit(res[0])
    else:
        sys.exit(res)



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import json
import os
import re
import shutil
import sys
import tempfile
from abc import (
    abstractmethod,
)

from packaging.version import (
    Version,
)

from . import (
    LOGGER,
)
from .constants import (
    DEFAULT_SDKCONFIG,
    IDF_PY,
    IDF_SIZE_PY,
    IDF_VERSION,
    PROJECT_DESCRIPTION_JSON,
)
from .manifest.manifest import (
    FolderRule,
    Manifest,
)
from .utils import (
    BuildError,
    dict_from_sdkconfig,
    find_first_match,
    rmdir,
    subprocess_run,
    to_list,
)

try:
    from typing import (
        Pattern,
        TextIO,
    )
except ImportError:
    pass


class App:
    TARGET_PLACEHOLDER = '@t'  # replace it with self.target
    WILDCARD_PLACEHOLDER = '@w'  # replace it with the wildcard, usually the sdkconfig
    NAME_PLACEHOLDER = '@n'  # replace it with self.name
    FULL_NAME_PLACEHOLDER = '@f'  # replace it with escaped self.app_dir
    INDEX_PLACEHOLDER = '@i'  # replace it with the build index

    BUILD_SYSTEM = 'unknown'

    SDKCONFIG_LINE_REGEX = re.compile(r"^([^=]+)=\"?([^\"\n]*)\"?\n*$")

    SIZE_JSON = 'size.json'

    # could be assigned later, used for filtering out apps by supported_targets
    MANIFEST = None  # type: Manifest | None

    # This RE will match GCC errors and many other fatal build errors and warnings as well
    LOG_ERROR_WARNING_REGEX = re.compile(r'(?:error|warning):', re.MULTILINE | re.IGNORECASE)
    # Log this many trailing lines from a failed build log, also
    LOG_DEBUG_LINES = 25
    # IGNORE_WARNING_REGEX is a regex for warnings to be ignored. Could be assigned later
    IGNORE_WARNS_REGEXES = []  # type: list[Pattern]

    def __init__(
        self,
        app_dir,
        target,
        sdkconfig_path=None,
        config_name=None,
        work_dir=None,
        build_dir='build',
        build_log_path=None,
        size_json_path=None,
        check_warnings=False,
        preserve=True,
    ):  # type: (str, str, str | None, str | None, str | None, str, str | None, str | None, bool, bool) -> None
        # These internal variables store the paths with environment variables and placeholders;
        # Public properties with similar names use the _expand method to get the actual paths.
        self._app_dir = app_dir
        self._work_dir = work_dir or app_dir
        self._build_dir = build_dir or 'build'
        self._build_log_path = build_log_path
        self._size_json_path = size_json_path

        self.name = os.path.basename(os.path.realpath(app_dir))
        self.sdkconfig_path = sdkconfig_path
        self.config_name = config_name
        self.target = target
        self.check_warnings = check_warnings
        self.preserve = preserve

        # Some miscellaneous build properties which are set later, at the build stage
        self.dry_run = False
        self.index = None
        self.verbose = False

    def __repr__(self):
        return '({}) App {}, target {}, sdkconfig {}, build in {}'.format(
            self.BUILD_SYSTEM,
            self.app_dir,
            self.target,
            self.sdkconfig_path or '(default)',
            self.build_path,
        )

    def __lt__(self, other):
        if self.app_dir != other.app_dir:
            return self.app_dir < other.app_dir
        elif self.target != other.target:
            return self.target < other.target
        else:
            return self.config_name < other.config_name

    def __eq__(self, other):
        for props in [
            'app_dir',
            'work_dir',
            'build_dir',
            'size_json_path',
            'sdkconfig_path',
            'config_name',
            'target',
            'check_warnings',
            'preserve',
        ]:
            if getattr(self, props) != getattr(other, props):
                return False

        return True

    def __hash__(self):
        return hash(
            (
                self._app_dir,
                self._work_dir,
                self._build_dir,
                self._build_log_path,
                self._size_json_path,
                self.name,
                self.sdkconfig_path,
                self.config_name,
                self.target,
                self.check_warnings,
                self.preserve,
            )
        )

    def _expand(self, path):  # type: (str) -> str
        """
        Internal method, expands any of the placeholders in {app,work,build} paths.
        """
        if not path:
            return path

        if self.index is not None:
            path = path.replace(self.INDEX_PLACEHOLDER, str(self.index))
        path = path.replace(self.TARGET_PLACEHOLDER, self.target)
        path = path.replace(self.NAME_PLACEHOLDER, self.name)
        if self.FULL_NAME_PLACEHOLDER in path:  # to avoid recursion to the call to app_dir in the next line:
            path = path.replace(self.FULL_NAME_PLACEHOLDER, self.app_dir.replace(os.path.sep, '_'))
        wildcard_pos = path.find(self.WILDCARD_PLACEHOLDER)
        if wildcard_pos != -1:
            if self.config_name:
                # if config name is defined, put it in place of the placeholder
                path = path.replace(self.WILDCARD_PLACEHOLDER, self.config_name)
            else:
                # otherwise, remove the placeholder and one character on the left
                # (which is usually an underscore, dash, or other delimiter)
                left_of_wildcard = max(0, wildcard_pos - 1)
                right_of_wildcard = wildcard_pos + len(self.WILDCARD_PLACEHOLDER)
                path = path[0:left_of_wildcard] + path[right_of_wildcard:]
        path = os.path.expandvars(path)
        return path

    @property
    def app_dir(self):
        """
        :return: directory of the app
        """
        return self._expand(self._app_dir)

    @property
    def work_dir(self):
        """
        :return: directory where the app should be copied to, prior to the build.
        """
        return self._expand(self._work_dir)

    @property
    def build_dir(self):
        """
        :return: build directory, either relative to the work directory (if relative path is used) or absolute path.
        """
        return self._expand(self._build_dir)

    @property
    def build_path(self):
        if os.path.isabs(self.build_dir):
            return self.build_dir

        return os.path.realpath(os.path.join(self.work_dir, self.build_dir))

    @property
    def build_log_path(self):
        if self._build_log_path:
            return os.path.join(self.build_path, self._expand(self._build_log_path))

        return None

    @property
    def size_json_path(self):
        if self._size_json_path:
            return os.path.join(self.build_path, self._expand(self._size_json_path))

        return None

    def build_prepare(self):  # type: () -> dict[str, str]
        LOGGER.debug('=> Preparing Folders')
        if self.work_dir != self.app_dir:
            if os.path.exists(self.work_dir):
                LOGGER.debug('==> Work directory %s exists, removing', self.work_dir)
                if not self.dry_run:
                    shutil.rmtree(self.work_dir)
            LOGGER.debug('==> Copying app from %s to %s', self.app_dir, self.work_dir)
            if not self.dry_run:
                shutil.copytree(self.app_dir, self.work_dir)

        if os.path.exists(self.build_path):
            LOGGER.debug('==> Build directory %s exists, removing', self.build_path)
            if not self.dry_run:
                shutil.rmtree(self.build_path)

        if not self.dry_run:
            os.makedirs(self.build_path)

        # Prepare the sdkconfig file, from the contents of sdkconfig.defaults (if exists) and the contents of
        # build_info.sdkconfig_path, i.e. the config-specific sdkconfig file.
        #
        # Note: the build system supports taking multiple sdkconfig.defaults files via SDKCONFIG_DEFAULTS
        # CMake variable. However here we do this manually to perform environment variable expansion in the
        # sdkconfig files.
        sdkconfig_defaults_list = [
            'sdkconfig.defaults',
            'sdkconfig.defaults.' + self.target,
        ]
        if self.sdkconfig_path:
            sdkconfig_defaults_list.append(self.sdkconfig_path)

        LOGGER.debug('=> Generating sdkconfig file')
        sdkconfig_file = os.path.join(self.work_dir, 'sdkconfig')
        if os.path.exists(sdkconfig_file):
            LOGGER.debug('==> Removing sdkconfig file: %s', sdkconfig_file)
            if not self.dry_run:
                os.unlink(sdkconfig_file)

        LOGGER.debug('==> Creating sdkconfig file: %s', sdkconfig_file)
        cmake_vars = {}
        if not self.dry_run:
            with open(sdkconfig_file, 'w') as f_out:
                for sdkconfig_name in sdkconfig_defaults_list:
                    sdkconfig_path = os.path.join(self.work_dir, sdkconfig_name)
                    if not sdkconfig_path or not os.path.exists(sdkconfig_path):
                        continue
                    LOGGER.debug('==> Appending %s to sdkconfig', sdkconfig_name)
                    with open(sdkconfig_path, 'r') as f_in:
                        for line in f_in:
                            if not line.endswith('\n'):
                                line += '\n'
                            if isinstance(self, CMakeApp):
                                m = self.SDKCONFIG_LINE_REGEX.match(line)
                                key = m.group(1) if m else None
                                if key in self.SDKCONFIG_TEST_OPTS:
                                    cmake_vars[key] = m.group(2)
                                    continue
                                if key in self.SDKCONFIG_IGNORE_OPTS:
                                    continue
                            f_out.write(os.path.expandvars(line))
        else:
            for sdkconfig_name in sdkconfig_defaults_list:
                sdkconfig_path = os.path.join(self.work_dir, sdkconfig_name)
                if not sdkconfig_path:
                    continue
                LOGGER.debug('==> Considering sdkconfig %s', sdkconfig_path)
                if not os.path.exists(sdkconfig_path):
                    continue
                LOGGER.debug('==> Appending %s to sdkconfig', sdkconfig_name)

        return cmake_vars

    @classmethod
    def _get_default_sdkconfig_target(cls, path):  # type: (str) -> str | None
        # check if there's CONFIG_IDF_TARGET in sdkconfig.defaults
        default_sdkconfig_target = None  # type: str | None
        default_sdkconfig = os.path.join(path, DEFAULT_SDKCONFIG)
        if os.path.isfile(default_sdkconfig):
            sdkconfig_dict = dict_from_sdkconfig(default_sdkconfig)
            if 'CONFIG_IDF_TARGET' in sdkconfig_dict:
                default_sdkconfig_target = sdkconfig_dict['CONFIG_IDF_TARGET']
                LOGGER.debug(
                    'CONFIG_IDF_TARGET is set in %s. Set enable build targets to %s only.',
                    default_sdkconfig,
                    default_sdkconfig_target,
                )

        return default_sdkconfig_target

    @classmethod
    def enable_build_targets(cls, path):  # type: (str) -> list[str]
        default_sdkconfig_target = cls._get_default_sdkconfig_target(path)
        if cls.MANIFEST:
            res = cls.MANIFEST.enable_build_targets(path, default_sdkconfig_target)
        else:
            if default_sdkconfig_target:
                res = [default_sdkconfig_target]
            else:
                res = FolderRule.DEFAULT_BUILD_TARGETS

        return res

    @classmethod
    def enable_test_targets(cls, path):  # type: (str) -> list[str]
        default_sdkconfig_target = cls._get_default_sdkconfig_target(path)

        if cls.MANIFEST:
            return cls.MANIFEST.enable_test_targets(path, default_sdkconfig_target)

        return []

    @classmethod
    def requires_components(cls, path):  # type: (str) -> list[str]
        if cls.MANIFEST:
            return cls.MANIFEST.requires_components(path)

        return []

    @property
    def supported_targets(self):
        return self.enable_build_targets(self.app_dir)

    @property
    def verified_targets(self):
        return self.enable_test_targets(self.app_dir)

    @abstractmethod
    def build(
        self,
        depends_on_components=None,  # type: list[str] | str | None
        check_component_dependencies=False,  # type: bool
    ):  # type: (...) -> bool
        pass

    def write_size_json(self):
        map_file = find_first_match('*.map', self.build_path)
        if not map_file:
            LOGGER.warning(
                '.map file not found. Cannot write size json to file: %s',
                self.size_json_path,
            )
            return

        subprocess_run(
            (
                [
                    sys.executable,
                    str(IDF_SIZE_PY),
                ]
                + (['--json'] if IDF_VERSION < Version('5.1') else ['--format', 'json'])
                + [
                    '-o',
                    self.size_json_path,
                    map_file,
                ]
            ),
            check=True,
        )
        LOGGER.info('=> Generated size info to %s', self.size_json_path)

    def collect_size_info(self, output_fs):  # type: (TextIO) -> None
        if not os.path.isfile(self.size_json_path):
            self.write_size_json()

        size_info_dict = {
            'app_name': self.name,
            'config_name': self.config_name,
            'target': self.target,
            'path': self.size_json_path,
        }
        output_fs.write(json.dumps(size_info_dict) + '\n')
        LOGGER.info('=> Recorded size info file path in %s', output_fs.name)

    def to_json(self):
        # keeping backward compatibility, only provide these stuffs
        return json.dumps(
            {
                'build_system': self.BUILD_SYSTEM,
                'app_dir': self.app_dir,
                'work_dir': self.work_dir,
                'build_dir': self.build_dir,
                'build_log_path': self.build_log_path,
                'sdkconfig': self.sdkconfig_path,
                'config': self.config_name,
                'target': self.target,
                'verbose': self.verbose,
                'preserve': self.preserve,
            }
        )

    def is_error_or_warning(self, line):  # type: (str) -> tuple[bool, bool]  # is_error_or_warning, is_ignored
        if not self.LOG_ERROR_WARNING_REGEX.search(line):
            return False, False

        is_ignored = False
        for ignored in self.IGNORE_WARNS_REGEXES:
            if re.search(ignored, line):
                is_ignored = True
                break

        return True, is_ignored


class CMakeApp(App):
    BUILD_SYSTEM = 'cmake'

    # If these keys are present in sdkconfig.defaults, they will be extracted and passed to CMake
    SDKCONFIG_TEST_OPTS = [
        'EXCLUDE_COMPONENTS',
        'TEST_EXCLUDE_COMPONENTS',
        'TEST_COMPONENTS',
    ]

    # These keys in sdkconfig.defaults are not propagated to the final sdkconfig file:
    SDKCONFIG_IGNORE_OPTS = ['TEST_GROUPS']

    # While ESP-IDF component CMakeLists files can be identified by the presence of 'idf_component_register' string,
    # there is no equivalent for the project CMakeLists files. This seems to be the best option...
    CMAKE_PROJECT_LINE = r'include($ENV{IDF_PATH}/tools/cmake/project.cmake)'

    def build(
        self,
        depends_on_components=None,  # type: list[str] | str | None
        check_component_dependencies=False,  # type: bool
    ):  # type: (...) -> bool
        cmake_vars = self.build_prepare()

        if self.build_log_path:
            LOGGER.info('=> Writing build log to %s', self.build_log_path)
            log_file = open(self.build_log_path, 'w')
        else:
            # delete manually later, used for tracking debugging info
            log_file = tempfile.NamedTemporaryFile('w', delete=False)

        depends_on_components = to_list(depends_on_components)
        if depends_on_components is not None and check_component_dependencies:
            reconfigure_args = [
                sys.executable,
                str(IDF_PY),
                '-B',
                self.build_path,
                '-C',
                self.work_dir,
                '-DIDF_TARGET=' + self.target,
                'reconfigure',
            ]
            subprocess_run(
                reconfigure_args, log_terminal=False if self.build_log_path else True, log_fs=log_file, check=True
            )

            with open(os.path.join(self.build_path, PROJECT_DESCRIPTION_JSON)) as fr:
                build_components = set(item for item in json.load(fr)['build_components'] if item)

            if not set(depends_on_components).intersection(set(build_components)):
                LOGGER.info(
                    '=> Skip building... app %s depends on components: %s, while current build based on component dependencies: %s',
                    self.app_dir,
                    build_components,
                    depends_on_components,
                )
                return False

        build_args = [
            sys.executable,
            str(IDF_PY),
            '-B',
            self.build_path,
            '-C',
            self.work_dir,
            '-DIDF_TARGET=' + self.target,
        ]
        if cmake_vars:
            for key, val in cmake_vars.items():
                build_args.append('-D{}={}'.format(key, val))
            if 'TEST_EXCLUDE_COMPONENTS' in cmake_vars and 'TEST_COMPONENTS' not in cmake_vars:
                build_args.append('-DTESTS_ALL=1')
            if 'CONFIG_APP_BUILD_BOOTLOADER' in cmake_vars:
                # In case if secure_boot is enabled then for bootloader build need to add `bootloader` cmd
                build_args.append('bootloader')

        build_args.append('build')

        if self.verbose:
            build_args.append('-v')

        if self.dry_run:
            LOGGER.debug('==> Skipping... (dry run)')
            return True

        old_idf_target_env = os.getenv('IDF_TARGET')
        os.environ['IDF_TARGET'] = self.target  # pass the cmake check

        returncode = subprocess_run(
            build_args,
            log_terminal=False if self.build_log_path else True,
            log_fs=log_file,
        )

        if old_idf_target_env is not None:
            os.environ['IDF_TARGET'] = old_idf_target_env  # revert it back
        log_file.close()

        # help debug
        has_unignored_warning = False
        with open(log_file.name) as f:
            lines = [line.rstrip() for line in f.readlines() if line.rstrip()]
            for line in lines:
                is_error_or_warning, ignored = self.is_error_or_warning(line)
                if is_error_or_warning:
                    if ignored:
                        LOGGER.info('[Ignored warning] %s', line)
                    else:
                        LOGGER.warning('%s', line)
                        has_unignored_warning = True

            if returncode != 0:
                # print last few lines to help debug
                LOGGER.error(
                    'Last %s lines from the build log "%s":',
                    self.LOG_DEBUG_LINES,
                    self.build_log_path,
                )
                for line in lines[-self.LOG_DEBUG_LINES :]:
                    LOGGER.error('%s', line)

        if returncode == 0 and self.size_json_path:
            self.write_size_json()

        if not self.preserve:
            LOGGER.info('=> Removing build directory %s', self.build_path)
            exclude_list = []
            if self.size_json_path:
                exclude_list.append(os.path.basename(self.size_json_path))
            if self.build_log_path:
                exclude_list.append(os.path.basename(self.build_log_path))

            rmdir(
                self.build_path,
                exclude_file_patterns=exclude_list,
            )

        if returncode != 0:
            raise BuildError('Build failed with exit code {}'.format(returncode))

        if has_unignored_warning:
            raise BuildError('Build succeeded with warnings')

        LOGGER.info('=> Build succeeded')
        return True

    @classmethod
    def is_app(cls, path):  # type: (str) -> bool
        cmakelists_path = os.path.join(path, 'CMakeLists.txt')
        if not os.path.exists(cmakelists_path):
            return False

        with open(cmakelists_path) as fr:
            cmakelists_file_content = fr.read()

        if not cmakelists_file_content:
            return False

        if cls.CMAKE_PROJECT_LINE not in cmakelists_file_content:
            return False

        return True



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import shutil

import pytest

from idf_build_apps.app import (
    CMakeApp,
)
from idf_build_apps.constants import (
    IDF_PATH,
)


@pytest.mark.skipif(not shutil.which('idf.py'), reason='idf.py not found')
class TestBuild:
    def test_build_hello_world(self, tmpdir, capsys):
        path = IDF_PATH / 'examples' / 'get-started' / 'hello_world'

        CMakeApp(str(path), 'esp32', work_dir=str(tmpdir / 'test')).build()

        captured = capsys.readouterr()
        assert 'Configuring done' in captured.out
        assert 'Project build complete.' in captured.out

    @pytest.mark.parametrize(
        'depends_on_components, check_component_dependencies, assert_build_done',
        [
            (None, True, True),
            ([], True, False),
            ([], False, True),
            ('fake', True, False),
            ('fake', False, True),
            ('soc', True, True),
            ('soc', False, True),
            (['soc', 'fake'], True, True),
        ],
    )
    def test_build_with_depends_on_components(
        self, tmpdir, capsys, depends_on_components, check_component_dependencies, assert_build_done
    ):
        path = IDF_PATH / 'examples' / 'get-started' / 'hello_world'

        CMakeApp(str(path), 'esp32', work_dir=str(tmpdir / 'test')).build(
            depends_on_components=depends_on_components,
            check_component_dependencies=check_component_dependencies,
        )

        captured = capsys.readouterr()
        assert 'Configuring done' in captured.out
        if assert_build_done:
            assert 'Project build complete.' in captured.out
        else:
            assert 'Project build complete.' not in captured.out


@pytest.mark.skipif(not shutil.which('idf.py'), reason='idf.py not found')
def test_idf_version_keywords_type():
    from idf_build_apps.constants import (
        IDF_VERSION_MAJOR,
        IDF_VERSION_MINOR,
        IDF_VERSION_PATCH,
    )

    assert isinstance(IDF_VERSION_MAJOR, int)
    assert isinstance(IDF_VERSION_MINOR, int)
    assert isinstance(IDF_VERSION_PATCH, int)



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import inspect
import os

from idf_build_apps.manifest.manifest import (
    Manifest,
)


def test_manifest(tmpdir):
    yaml_file = tmpdir / 'test.yml'
    with open(yaml_file, 'w') as fw:
        fw.write(
            inspect.cleandoc(
                '''
            test1:
                enable:
                    - if: IDF_TARGET == "esp32" or IDF_TARGET == "esp32c3"
                    - if: IDF_TARGET == "esp32s2" and IDF_VERSION_MAJOR in [4, 5]
                disable_test:
                    - if: IDF_TARGET == "esp32c3"
        '''
            )
        )

    os.chdir(tmpdir)
    Manifest.ROOTPATH = tmpdir
    manifest = Manifest.from_file(yaml_file)

    assert manifest.enable_build_targets('test1') == ['esp32', 'esp32c3', 'esp32s2']
    assert manifest.enable_test_targets('test1') == ['esp32', 'esp32s2']



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['idf_build_apps', 'idf_build_apps.manifest']

package_data = \
{'': ['*']}

install_requires = \
['pyparsing', 'pyyaml', 'packaging']

extras_require = \
{":python_version < '3.4'": ['pathlib'],
 'doc': ['sphinx',
         'sphinx-rtd-theme',
         'sphinx_copybutton',
         'myst-parser',
         'sphinxcontrib-mermaid'],
 'test': ['pytest', 'pytest-cov']}

setup(name='idf-build-apps',
      version='0.4.1',
      description='Tools for building ESP-IDF related apps.',
      author=None,
      author_email='Fu Hanxi <fuhanxi@espressif.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
     )



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import logging
from pathlib import (
    Path,
)

from pyparsing import (
    CaselessLiteral,
    Char,
    Combine,
    Group,
    Literal,
    OneOrMore,
    Optional,
    ParseException,
    ParserElement,
    QuotedString,
    Word,
    alphas,
    hexnums,
    nums,
)

from ..constants import (
    ALL_TARGETS,
    IDF_PATH,
)


def get_defines(header_path):  # type: (Path) -> list[str]
    defines = []
    logging.debug('Reading macros from %s...', header_path)
    with open(str(header_path), 'r') as f:
        output = f.read()

    for line in output.split('\n'):
        line = line.strip()
        if len(line):
            defines.append(line)

    return defines


def parse_define(define_line):  # type: (str) -> ParserElement
    # Group for parsing literal suffix of a numbers, e.g. 100UL
    literal_symbol = Group(CaselessLiteral('L') | CaselessLiteral('U'))
    literal_suffix = OneOrMore(literal_symbol)

    # Define name
    name = Word(alphas, alphas + nums + '_')

    # Define value, either a hex, int or a string
    hex_value = Combine(Literal('0x') + Word(hexnums) + Optional(literal_suffix).suppress())('hex_value')
    int_value = Word(nums)('int_value') + ~Char('.') + Optional(literal_suffix)('literal_suffix')
    str_value = QuotedString('"')('str_value')

    # Remove optional parenthesis around values
    value = Optional('(').suppress() + (hex_value ^ int_value ^ str_value)('value') + Optional(')').suppress()

    expr = '#define' + Optional(name)('name') + Optional(value)
    res = expr.parseString(define_line)

    return res


class SocHeader(dict):
    def __init__(self, target):  # type: (str) -> None
        if target != 'linux':
            soc_header_dict = self._parse_soc_header(target)
        else:
            soc_header_dict = {}

        super(SocHeader, self).__init__(**soc_header_dict)

    @staticmethod
    def _parse_soc_header(target):  # type: (str) -> dict[str, any]
        soc_headers_dir_candidates = [
            # other branches
            IDF_PATH / 'components' / 'soc' / target / 'include' / 'soc',
            # release/v4.2
            IDF_PATH / 'components' / 'soc' / 'soc' / target / 'include' / 'soc',
        ]

        # get the soc_headers_dir
        soc_headers_dir = None
        for d in soc_headers_dir_candidates:
            if not d.is_dir():
                logging.debug('No soc header files folder: %s', d.absolute())
            else:
                soc_headers_dir = d
                break

        if not soc_headers_dir:
            return {}

        output_dict = {}
        for soc_header_file in soc_headers_dir.glob('*_caps.h'):
            for line in get_defines(soc_header_file):
                try:
                    res = parse_define(line)
                except ParseException:
                    logging.debug('Failed to parse: %s', line)
                    continue

                if 'str_value' in res:
                    output_dict[res.name] = res.str_value
                elif 'int_value' in res:
                    output_dict[res.name] = int(res.int_value)
                elif 'hex_value' in res:
                    output_dict[res.name] = int(res.hex_value, 16)

        return output_dict


SOC_HEADERS = {target: SocHeader(target) for target in ALL_TARGETS}
